/*
COPYRIGHT (C) 2003  Lorenzo Dozio (dozio@aero.polimi.it)
                    Paolo Mantegazza (mantegazza@aero.polimi.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/io.h>
#include <sys/poll.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "controlpanel.h"
#include "targetpanel.h"
#include "parameterspanel.h"
#include "scopepanel.h"
#include "logpanel.h"
#include "rtailab.h"
#include "scopecanvas.h"
#include "scopewindow.h"

Fl_Preferences app(Fl_Preferences::USER, "rtai.org", "RTAI-Lab" );

ControlPanel *CommandManager;
TargetPanel *TargetIDs;
ParametersPanel *ParametersManager;
ScopePanel *ScopeManager;
LogPanel *LogManager;

static int IsTargetConnected;
static unsigned int IsTargetRunning;
int EndApp = 0;
int EndAll = 0;
int NumTunableParameters = 0;
int NumRtScopes = 0;
int NumRtLogs = 0;

RT_TASK *rt_TargetInterfaceTask;
pthread_t rt_TargetInterfaceThread;
pthread_t rt_GetScopeDataThread[NSCOPETHREADS];
pthread_t rt_GetLogDataThread[NLOGTHREADS];
pthread_t rt_GetTargetTimeThread;

char *TargetScopeMbxID;
char *TargetLogMbxID;

rtTargetParamInfo *TunableParameters;
rtTargetScopeInfo RtScopes[MAX_RTAI_SCOPES];
rtTargetLogInfo RtLogs[MAX_RTAI_LOGS];
rtBatchDownload BatchParameters[MAX_BATCH_PARAMS];

static float TargetTime;
int *scpWinRedraw;
static long TargetNode = 0;

unsigned long get_an_id(const char *root)
{
	int i;
	char name[7];
	for (i = 0; i < MAX_NHOSTS; i++) {
		sprintf(name, "%s%d", root, i);
		if (!rt_get_adr(nam2num(name))) {
			return nam2num(name);
		}
	}
	return 0;
}

static volatile int GlobalRet[16];

static inline void RT_RPC(RT_TASK *task, unsigned int msg, unsigned int *reply)
{
	GlobalRet[msg & 0xf] = 0;
	rt_send(task, msg);
	while (!GlobalRet[msg & 0xf]) {
		Fl::wait(FLTK_EVENTS_TICK);
	}
}

static inline void RT_RETURN(RT_TASK *task, unsigned int reply)
{
	GlobalRet[reply] = 1;
}

static void *rt_GetTargetTime(void *arg)
{
	RT_TASK *rt_GetTargetTimeTask;
	int n;

	rt_allow_nonroot_hrt();
	if (!(rt_GetTargetTimeTask = rt_task_init_schmod(get_an_id("HGT"), rt_GetTargetTimePriority, 0, 0, SCHED_FIFO, 0xFF))) {
		printf("Cannot init rt_GetTargetTimeTask\n");
		return (void *)1;
	}
	mlockall(MCL_CURRENT | MCL_FUTURE);

	while (true) {
		if (EndApp || !IsTargetConnected) break;
		rt_send_if(rt_TargetInterfaceTask, GET_TARGET_TIME);
		for (n = 0; n < NumRtScopes; n++) {
			scpWinRedraw[n] = true;
		}
		msleep(400);
	}

	rt_task_delete(rt_GetTargetTimeTask);

	return 0;
}

static void *rt_GetLogData(void *arg)
{
	RT_TASK *rt_GetLogDataTask;
	MBX *rt_GetLogDataMbx;
	char rt_GetLogDataMbxName[7];
	long rt_GetLogDataPort;
	int index = *((int *)arg);
	float MsgBuf[MAX_MSG_LEN/sizeof(float)];
	int MsgLen, MaxMsgLen, MsgData, DataBytes;
	int n, i, j, k;
	int DataCnt = 0;

	rt_allow_nonroot_hrt();
	if (!(rt_GetLogDataTask = rt_task_init_schmod(get_an_id("HGL"), rt_GetLogPriority, 0, 0, SCHED_RR, 0xFF))) {
		printf("Cannot init Host GetLogData Task\n");
		return (void *)1;
	}
	rt_GetLogDataPort = rt_request_port(TargetNode);
	sprintf(rt_GetLogDataMbxName, "%s%d", TargetLogMbxID, index);
	if (!(rt_GetLogDataMbx = (MBX *)RT_get_adr(TargetNode, rt_GetLogDataPort, rt_GetLogDataMbxName))) {
		printf("Error in getting %s mailbox address\n", rt_GetLogDataMbxName);
		exit(1);
	}
	DataBytes = RtLogs[index].nrow*RtLogs[index].ncol*sizeof(float);
	MaxMsgLen = (MAX_MSG_LEN/DataBytes)*DataBytes;

	rt_send(rt_TargetInterfaceTask, 0);
	mlockall(MCL_CURRENT | MCL_FUTURE);

	while (true) {
		if (EndApp || !IsTargetConnected) break;
		MsgLen = (((int)(DataBytes*REFRESH_RATE*(1./RtLogs[index].dt)))/DataBytes)*DataBytes;
		if (MsgLen < DataBytes) {
			MsgLen = DataBytes;
		}
		if (MsgLen > MaxMsgLen) {
			MsgLen = MaxMsgLen;
		}
		while (RT_mbx_receive_if(TargetNode, rt_GetLogDataPort, rt_GetLogDataMbx, &MsgBuf, MsgLen)) {
			if (!IsTargetConnected) goto end;
			msleep(10);
		}
		MsgData = MsgLen/DataBytes;
		for (n = 0; n < MsgData; n++) {
			if (LogManager) {
				if (LogManager->IsLoggingEnabled(index)) {
					fprintf(LogManager->LogFile[index], "Data # %d\n", ++DataCnt);
					for (i = 0; i < RtLogs[index].nrow; i++) {
                                        	j = n*RtLogs[index].nrow*RtLogs[index].ncol + i;
                                        	for (k = 0; k < RtLogs[index].ncol; k++) {
                                        	        fprintf(LogManager->LogFile[index], "%1.5f ", MsgBuf[j]);
                                        	        j += RtLogs[index].nrow;
                                        	}
						fprintf(LogManager->LogFile[index], "\n");
					}
				}
			}
		}
	}

end:
	rt_release_port(TargetNode, rt_GetLogDataPort);
	rt_task_delete(rt_GetLogDataTask);

	return 0;
}

static void *rt_GetScopeData(void *arg)
{
	RT_TASK *rt_GetScopeDataTask;
	MBX *rt_GetScopeDataMbx;
	char rt_GetScopeDataMbxName[7];
	long rt_GetScopeDataPort;
	ScopeWindow *ScopeW;
	int MsgData = 0, n, nn;
	int index = *((int *)arg);
	int jscope, jlog;
	float MsgBuf[MAX_MSG_LEN/sizeof(float)];
	int MsgLen, MaxMsgLen, TracesBytes;
	int ScopeFrameExist = false;

	rt_allow_nonroot_hrt();
	if (!(rt_GetScopeDataTask = rt_task_init_schmod(get_an_id("HGS"), rt_GetScopePriority, 0, 0, SCHED_RR, 0xFF))) {
		printf("Cannot init Host GetScopeData Task\n");
		return (void *)1;
	}
	rt_GetScopeDataPort = rt_request_port(TargetNode);
	sprintf(rt_GetScopeDataMbxName, "%s%d", TargetScopeMbxID, index);
	if (!(rt_GetScopeDataMbx = (MBX *)RT_get_adr(TargetNode, rt_GetScopeDataPort, rt_GetScopeDataMbxName))) {
		printf("Error in getting %s mailbox address\n", rt_GetScopeDataMbxName);
		exit(1);
	}
	TracesBytes = (RtScopes[index].ntraces + 1)*sizeof(float);
//	TracesBytes = RtScopes[index].ntraces*sizeof(float);
	MaxMsgLen = (MAX_MSG_LEN/TracesBytes)*TracesBytes;
	ScopeW = new ScopeWindow(index, RtScopes[index].name, RtScopes[index].ntraces, (double)(1./RtScopes[index].dt));
	ScopeW->Scope->setTracePointer(0);

	rt_send(rt_TargetInterfaceTask, 0);
	mlockall(MCL_CURRENT | MCL_FUTURE);

	while (true) {
		if (EndApp || !IsTargetConnected) break;
		MsgLen = (((int)(TracesBytes*REFRESH_RATE*ScopeW->Scope->SamplingFrequency()))/TracesBytes)*TracesBytes;
		if (MsgLen < TracesBytes) {
			MsgLen = TracesBytes;
		}
		if (MsgLen > MaxMsgLen) {
			MsgLen = MaxMsgLen;
		}
		while (RT_mbx_receive_if(TargetNode, rt_GetScopeDataPort, rt_GetScopeDataMbx, &MsgBuf, MsgLen)) {
			if (!IsTargetConnected) goto end;
			if (RtScopes[index].visible) {
				if (!ScopeFrameExist) {
					Fl::lock();
					ScopeW->show();
					ScopeFrameExist = true;
					Fl::unlock();
				}
			} else {
				if (ScopeFrameExist) {
					Fl::lock();
					ScopeW->hide();
					ScopeFrameExist = false;
					Fl::unlock();
				}
			}
			msleep(10);
		}
		MsgData = MsgLen/TracesBytes;
		Fl::lock();
		jscope = 1;
//		jscope = 0;
		jlog = 0;
		for (n = 0; n < MsgData; n++) {
			if (ScopeW->Scope->TracePointer() >= ScopeW->Scope->TraceLength()) {
				if (ScopeW->Scope->OneShotFlag == true) {
					ScopeW->Scope->StopRedraw = true;
					goto plot;
				} else {
					ScopeW->Scope->StopRedraw = false;
					ScopeW->Scope->setTracePointer(0);
				}
			}
			for (nn = 0; nn < RtScopes[index].ntraces - 1; nn++) {
				ScopeW->Scope->fillTrace(ScopeW->Scope->TracePointer(), nn, MsgBuf[jscope++]);
			}
			ScopeW->Scope->fillTrace(ScopeW->Scope->incrementTracePointer(), nn, MsgBuf[jscope++]);
			if (ScopeManager) {
				if (ScopeManager->IsLoggingEnabled(index)) {
					for (nn = 0; nn < RtScopes[index].ntraces + 1; nn++) {
						fprintf(ScopeManager->LogFile[index], "%1.5f ", MsgBuf[jlog++]);
					}
					fprintf(ScopeManager->LogFile[index], "\n");
				}
			}
			jscope++;
		}
plot:
		if (ScopeFrameExist) {
			if (!ScopeW->Scope->StopRedraw && !ScopeW->Scope->ScopePauseFlag) {
				ScopeW->Scope->redraw();
			}
			if (scpWinRedraw[index]) {
				ScopeW->getTargetTime(TargetTime);
				ScopeW->scpWin->redraw();
				scpWinRedraw[index] = false;
			}
		}
		Fl::unlock();
	}

end:
	ScopeW->hide();
	delete(ScopeW);
	rt_release_port(TargetNode, rt_GetScopeDataPort);
	rt_task_delete(rt_GetScopeDataTask);

	return 0;
}

static void *rt_TargetInterface(void *args)
{
	unsigned int code;
	char *TargetIP;
	char *TargetInterfaceTaskName;
	struct sockaddr_in TargetAddr;
	int ConnectionPollingCounter = 0;
	int ConnectionPollingMax = 5;
	long TargetPort = 0;
	RT_TASK *IfTask = NULL, *task;
	rtTargetParamInfo TargetParameters;

	rt_allow_nonroot_hrt();
	if (!(rt_TargetInterfaceTask = rt_task_init_schmod(get_an_id("HTI"), rt_TargetInterfacePriority, 0, 0, SCHED_FIFO, 0xFF))) {
		printf("Cannot init rt_TargetInterfaceTask\n");
		exit(1);
	}

	while (!EndApp) {

		if (!(task = rt_receive(0, &code))) continue;

		switch (code & 0xffff) {

			case CONNECT_TO_TARGET:
				Fl::lock();
				if (!TargetIDs) {
					TargetIDs = new TargetPanel();
				}
				TargetIP = TargetIDs->getIP();
				TargetInterfaceTaskName = TargetIDs->getIfTaskID();
				TargetScopeMbxID = TargetIDs->getScopeID();
				TargetLogMbxID = TargetIDs->getLogID();
				Fl::unlock();

				if (!strcmp(TargetIP, "0")) {
					TargetNode = 0;
					Fl::lock();
					CommandManager->activate_Connect();
					Fl::unlock();
				} else {
					char sbuf[100];
					sprintf(sbuf, "Trying to connect to %s", TargetIP);
					Fl::lock();
					CommandManager->MessageValue(strdup(sbuf));
					CommandManager->ConnectLabel("Connecting...");
					CommandManager->deactivate_Connect();
					Fl::unlock();
					inet_aton(TargetIP, &TargetAddr.sin_addr);
					TargetNode = TargetAddr.sin_addr.s_addr;
					while ((TargetPort = rt_request_port(TargetNode)) <= 0 && ConnectionPollingCounter++ <= ConnectionPollingMax) {
						msleep(100);
					}
					Fl::lock();
					CommandManager->activate_Connect();
					Fl::unlock();
					if (ConnectionPollingCounter >= ConnectionPollingMax) {
						ConnectionPollingCounter = 0;
						Fl::lock();
						CommandManager->ConnectLabel("Connect");
						CommandManager->MessageValue(" Sorry, no route to target");
						Fl::unlock();
						RT_RETURN(task, CONNECT_TO_TARGET);
						break;
					}
				}
				if (!(IfTask = (RT_TASK *)RT_get_adr(TargetNode, TargetPort, TargetInterfaceTaskName))) {
					Fl::lock();
					CommandManager->MessageValue(" Error in getting TargetInterfaceTask address");
					CommandManager->ConnectLabel("Connect");
					Fl::unlock();
					RT_RETURN(task, CONNECT_TO_TARGET);
					break;
				}
				{
					unsigned int Request = 'c';
					RT_rpc(TargetNode, TargetPort, IfTask, Request, &IsTargetRunning);
					NumTunableParameters = IsTargetRunning & 0xffff;
					IsTargetRunning >>= 16;
					TunableParameters = new rtTargetParamInfo[NumTunableParameters];
				}
				{
					char Request = 'i';
					for (int n = 0; n < NumTunableParameters; n++) {
						RT_rpcx(TargetNode, TargetPort, IfTask, &Request, &TargetParameters, sizeof(char), sizeof(TargetParameters));
						strncpyz(TunableParameters[n].modelName, TargetParameters.modelName, MAX_NAME_SIZE);
						strncpyz(TunableParameters[n].blockName, TargetParameters.blockName, MAX_NAME_SIZE);
						strncpyz(TunableParameters[n].paramName, TargetParameters.paramName, MAX_NAME_SIZE);
						TunableParameters[n].dataType  = TargetParameters.dataType;
						TunableParameters[n].dataClass = TargetParameters.dataClass;
						TunableParameters[n].dataValue = TargetParameters.dataValue;
					}
				}
				{
					int n;
					for (n = 0; n < MAX_RTAI_SCOPES; n++) {
						char name[7];
						char ScopeName[MAX_NAME_SIZE];
						sprintf(name, "%s%d", TargetScopeMbxID, n);
						if (!RT_get_adr(TargetNode, TargetPort, name)) break;
						RtScopes[n].visible = false;
						RT_rpcx(TargetNode, TargetPort, IfTask, &n, &RtScopes[n].ntraces, sizeof(int), sizeof(int));
						RT_rpcx(TargetNode, TargetPort, IfTask, &n, &ScopeName, sizeof(int), sizeof(ScopeName));
						strncpyz(RtScopes[n].name, ScopeName, MAX_NAME_SIZE);
						RT_rpcx(TargetNode, TargetPort, IfTask, &n, &RtScopes[n].dt, sizeof(int), sizeof(float));
					}
					NumRtScopes = n;
					scpWinRedraw = new int[NumRtScopes];
				}
				{
					int Request = -1, Msg;
					RT_rpcx(TargetNode, TargetPort, IfTask, &Request, &Msg, sizeof(int), sizeof(int));
				}
				{
					int n;
					for (n = 0; n < MAX_RTAI_LOGS; n++) {
						char name[7];
						char LogName[MAX_NAME_SIZE];
						sprintf(name, "%s%d", TargetLogMbxID, n);
						if (!RT_get_adr(TargetNode, TargetPort, name)) break;
						RT_rpcx(TargetNode, TargetPort, IfTask, &n, &RtLogs[n].nrow, sizeof(int), sizeof(int));
						RT_rpcx(TargetNode, TargetPort, IfTask, &n, &RtLogs[n].ncol, sizeof(int), sizeof(int));
						RT_rpcx(TargetNode, TargetPort, IfTask, &n, &LogName, sizeof(int), sizeof(LogName));
						strncpyz(RtLogs[n].name, LogName, MAX_NAME_SIZE);
						RT_rpcx(TargetNode, TargetPort, IfTask, &n, &RtLogs[n].dt, sizeof(int), sizeof(float));
					}
					NumRtLogs = n;
				}
				{
					int Request = -1, Msg;
					RT_rpcx(TargetNode, TargetPort, IfTask, &Request, &Msg, sizeof(int), sizeof(int));
				}
				IsTargetConnected = 1;
				for (int n = 0; n < NumRtScopes; n++) {
					unsigned int msg;
					pthread_create(&rt_GetScopeDataThread[n], NULL, rt_GetScopeData, &n);
					rt_receive(0, &msg);
				}
				for (int n = 0; n < NumRtLogs; n++) {
					unsigned int msg;
					pthread_create(&rt_GetLogDataThread[n], NULL, rt_GetLogData, &n);
					rt_receive(0, &msg);
				}
				pthread_create(&rt_GetTargetTimeThread, NULL, rt_GetTargetTime, NULL);
				{ 
					char mname[100];
					sprintf(mname, " Connected to %s", TunableParameters[0].modelName);
					Fl::lock();
					CommandManager->activate_Start();
					CommandManager->activate_ChangeParams();
					CommandManager->activate_RtPlots();
					CommandManager->activate_RtLogs();
					if (IsTargetRunning) {
						CommandManager->StartLabel("Stop real-time code");
					} else {
						CommandManager->StartLabel("Start real-time code");
					}
					CommandManager->ConnectLabel("Disconnect");
					CommandManager->MessageValue(strdup(mname));
					Fl::unlock();
				}
				RT_RETURN(task, CONNECT_TO_TARGET);
				break;

			case DISCONNECT_FROM_TARGET:
				ConnectionPollingCounter = 0;
				IsTargetConnected = 0;
				Fl::lock();
				CommandManager->deactivate_RtPlots();
				CommandManager->deactivate_RtLogs();
				CommandManager->deactivate_ChangeParams();
				CommandManager->deactivate_Start();
				if (ScopeManager) {
					ScopeManager->hide();
				}
				if (LogManager) {
					LogManager->hide();
				}
				CommandManager->ConnectLabel("Connect");
				CommandManager->MessageValue(" Disconnected");
				Fl::unlock();
				rt_release_port(TargetNode, TargetPort);
				TargetPort = 0;				
				free(TunableParameters);
				RT_RETURN(task, DISCONNECT_FROM_TARGET);
				break;

			case START_TARGET:
				if (!IsTargetRunning) {
					unsigned int Request = 's';
					RT_rpc(TargetNode, TargetPort, IfTask, Request, &IsTargetRunning);
					CommandManager->StartLabel("Stop real-time code");
				} else {
					CommandManager->StartLabel("Start real-time code");
				}
				RT_RETURN(task, START_TARGET);
				break;
			
			case STOP_TARGET:
				if (IsTargetRunning) {
					unsigned int Request = 't';
					IsTargetConnected = 0;
					for (int n = 0; n < NumRtScopes; n++) {
						pthread_join(rt_GetScopeDataThread[n], NULL);
					}
					for (int n = 0; n < NumRtLogs; n++) {
						pthread_join(rt_GetLogDataThread[n], NULL);
					}
					pthread_join(rt_GetTargetTimeThread, NULL);
					RT_rpc(TargetNode, TargetPort, IfTask, Request, &IsTargetRunning);
					rt_release_port(TargetNode, TargetPort);
					ConnectionPollingCounter = 0;
					TargetNode = 0;
					TargetPort = 0;	
					Fl::lock();
					CommandManager->deactivate_RtPlots();
					CommandManager->deactivate_RtLogs();
					CommandManager->deactivate_ChangeParams();
					CommandManager->deactivate_Start();
					CommandManager->ConnectLabel("Connect");
					CommandManager->MessageValue(" Disconnected");
					Fl::unlock();
				}
				RT_RETURN(task, STOP_TARGET);
				break;

			case UPDATE_PARAM: {
				unsigned int Request = 'p';
				int mapOffset = code >> 16;
				int isParamUpdated;
				double newVal = (double)TunableParameters[mapOffset].dataValue;
				RT_rpc(TargetNode, TargetPort, IfTask, Request, &IsTargetRunning);
				RT_rpcx(TargetNode, TargetPort, IfTask, &mapOffset, &isParamUpdated, sizeof(int), sizeof(int));
				RT_rpcx(TargetNode, TargetPort, IfTask, &newVal, &isParamUpdated, sizeof(double), sizeof(int));
				RT_RETURN(task, UPDATE_PARAM);
				break;
			}

			case BATCH_DOWNLOAD: {
				unsigned int Request = 'd';
				int isParamUpdated;
				int cnt = ParametersManager->BatchCounter();
				RT_rpc(TargetNode, TargetPort, IfTask, Request, &IsTargetRunning);
				RT_rpcx(TargetNode, TargetPort, IfTask, &cnt, &isParamUpdated, sizeof(int), sizeof(int));
				RT_rpcx(TargetNode, TargetPort, IfTask, &BatchParameters, &isParamUpdated, sizeof(rtBatchDownload)*cnt, sizeof(int));
				RT_RETURN(task, BATCH_DOWNLOAD);
				break;
			}

			case GET_PARAMS: {
				unsigned int Request = 'g';
				int n;
				RT_rpc(TargetNode, TargetPort, IfTask, Request, &IsTargetRunning);
				for (n = 0; n < NumTunableParameters; n++) {
					RT_rpcx(TargetNode, TargetPort, IfTask, &Request, &TunableParameters[n].dataValue, sizeof(int), sizeof(double));
				}
				ParametersManager->GetParameters();
				RT_RETURN(task, GET_PARAMS);
				break;
			}

			case GET_TARGET_TIME: {
				unsigned int Request = 'm';
				int Msg;
				RT_rpc(TargetNode, TargetPort, IfTask, Request, &IsTargetRunning);
				RT_rpcx(TargetNode, TargetPort, IfTask, &Msg, &TargetTime, sizeof(int), sizeof(float));
				break;
			}

			case CLOSE:
				rt_task_delete(rt_TargetInterfaceTask);
				return 0;

			default:
				break;
		}
	}

	rt_task_delete(rt_TargetInterfaceTask);

	return 0;
}

void cb_batch_update(Fl_Widget *o, void *v)
{
	int i, n;
	time_t curtime = time(NULL);
	struct tm *loctime = localtime(&curtime);

	for (i = n = 0; i < ParametersManager->BatchCounter(); i++) {
		n += ParametersManager->UpdateParameter(BatchParameters[i].index, BatchParameters[i].value, loctime);
	}
	if (ParametersManager->BatchCounter() > 0 && n == ParametersManager->BatchCounter()) {
		RT_RPC(rt_TargetInterfaceTask, BATCH_DOWNLOAD, 0);
	}
}

void cb_update(Fl_Widget *o, void *v)
{
	int n = (int)v;
	double value = atof(((Fl_Input*)o)->value());
	time_t curtime = time(NULL);
	struct tm *loctime = localtime(&curtime);

	if (ParametersManager->IsBatch() && ParametersManager->BatchCounter() < MAX_BATCH_PARAMS) {
		BatchParameters[ParametersManager->BatchCounter()].index = n;
		BatchParameters[ParametersManager->increment_BatchCounter()].value = value;
	} else {
		if (ParametersManager->UpdateParameter(n, value, loctime)) {
			RT_RPC(rt_TargetInterfaceTask, (n << 16) | UPDATE_PARAM, 0);
		}
	}
}

void cb_get_parameters(Fl_Widget *, void *)
{
	RT_RPC(rt_TargetInterfaceTask, GET_PARAMS, 0);
}

void cb_connect(Fl_Widget *, void *)
{
	IsTargetConnected ? RT_RPC(rt_TargetInterfaceTask, DISCONNECT_FROM_TARGET, 0) : RT_RPC(rt_TargetInterfaceTask, CONNECT_TO_TARGET, 0);
}

void cb_start(Fl_Widget *, void *)
{
	IsTargetRunning ? RT_RPC(rt_TargetInterfaceTask, STOP_TARGET, 0) : RT_RPC(rt_TargetInterfaceTask, START_TARGET, 0);
}

int main(int argc, char **argv)
{
	RT_TASK *MainTask;

	rt_allow_nonroot_hrt();
	if (!(MainTask = rt_task_init_schmod(get_an_id("HMN"), rt_MainPriority, 0, 0, SCHED_FIFO, 0xFF))) {
		printf("Cannot init HostMainTask\n");
		return 1;
	}
	CommandManager = new ControlPanel();
	CommandManager->show();
	Fl::lock();
	pthread_create(&rt_TargetInterfaceThread, NULL, rt_TargetInterface, NULL);
	while (!EndAll) {
		Fl::wait(FLTK_EVENTS_TICK);
	}
	rt_task_delete(MainTask);
	return 0;
}
